package geniusweb.voting;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;
import java.util.stream.Collectors;

import geniusweb.actions.PartyId;
import geniusweb.actions.Vote;
import geniusweb.actions.Votes;
import geniusweb.issuevalue.Bid;
import tudelft.utilities.immutablelist.FixedList;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.PowerSet;

/**
 * Utility class. Typically contains all the votes collected from 1 round. Can
 * find the best combinations of votes, eg to get the maximum consensus group
 */
public class CollectedVotes {
	private final Map<PartyId, Votes> allvotes;
	private final Map<PartyId, Integer> powers;

	// caching
	private transient Map<Bid, Set<Vote>> allBids = null;

	/**
	 * 
	 * @param votesMap the {@link Votes} done for each party
	 * @param power    the power of each party. The power is an integer number.
	 *                 In the voting process, the party powers add up to form
	 *                 the total voting power. This set may contain parties that
	 *                 are not in newmap.
	 */
	public CollectedVotes(Map<PartyId, Votes> votesMap,
			Map<PartyId, Integer> power) {
		this.allvotes = new HashMap<>(votesMap);
		this.powers = new HashMap<>(power);
		if (!powers.keySet().containsAll(allvotes.keySet())) {
			throw new IllegalArgumentException(
					"All voting parties must have a power");
		}
	}

	/**
	 * @param votes additional set of votes for a new party that did not yet
	 *              vote.
	 * @param power powers for the new party, can be null if power already set
	 *              previously.
	 * @return new CollectedVotes with the new votes added.
	 */
	public CollectedVotes with(Votes votes, Integer power) {
		PartyId actor = votes.getActor();
		if (allvotes.containsKey(actor))
			throw new IllegalArgumentException(
					"Party " + actor + " already voted");

		Map<PartyId, Integer> newpowers = new HashMap<>(powers);
		if (!powers.containsKey(actor)) {
			if (power == null)
				throw new IllegalArgumentException(
						"Power must be set for new actor " + actor);
			newpowers.put(actor, power);
		}

		Map<PartyId, Votes> newmap = new HashMap<>(allvotes);
		newmap.put(actor, votes);
		return new CollectedVotes(newmap, newpowers);
	}

	/**
	 * @return all the votes placed so far
	 */
	public Map<PartyId, Votes> getVotes() {
		return Collections.unmodifiableMap(allvotes);
	}

	/**
	 * @return the powers of all parties
	 */
	public Map<PartyId, Integer> getPowers() {
		return Collections.unmodifiableMap(powers);
	}

	/**
	 * @return a map containing for each {@link Bid}s a subset of
	 *         {@link PartyId}s that placed a satisfied vote for that bid, such
	 *         that the subset has maximum power. Maximum power means that there
	 *         is no other subset of satisfied votes for that bid with more
	 *         group power. There may be other subsets of equal power for the
	 *         bid. If there are no concensus possibilities at all for a bid,
	 *         the bid is not placed in the map.
	 */
	public Map<Bid, Set<PartyId>> getMaxAgreements() {
		Map<Bid, Set<PartyId>> agreements = new HashMap<>();
		Map<Bid, Set<Vote>> allbidsmap = getAllBids();
		for (Bid bid : allbidsmap.keySet()) {
			Set<PartyId> max = getMaxPowerGroup(allbidsmap.get(bid));
			if (!max.isEmpty())
				agreements.put(bid, max);
		}
		return agreements;
	}

	/**
	 * @return all bids that were voted on and the votes for that bid. This
	 *         basically reverses keys and values in {@link #allvotes}
	 */
	public Map<Bid, Set<Vote>> getAllBids() {
		if (allBids == null)
			allBids = getAllBids1();
		return allBids;
	}

	/**
	 * 
	 * @param parties the parties to be removed
	 * @return new CollectedVotes without given parties
	 */
	public CollectedVotes without(Set<PartyId> parties) {
		Map<PartyId, Votes> newvotes = new HashMap<>(allvotes);
		newvotes.keySet().removeAll(parties);
		Map<PartyId, Integer> newpower = new HashMap<>(powers);
		newpower.keySet().removeAll(parties);
		return new CollectedVotes(newvotes, newpower);
	}

	/**
	 * 
	 * @param parties a list of parties
	 * @return total voting power of the parties
	 */
	public Integer getTotalPower(Set<PartyId> parties) {
		return parties.stream().map(p -> powers.get(p)).reduce(0, Integer::sum);
	}

	/**
	 * Breath-first optimized search, similar as {@link #getMaxPowerGroup(Set)}
	 * but more efficient on average (remains to be proven). Notice that this
	 * algorithm may be more expensive particularly if there is no consensus at
	 * all, because of the added overhead of the breath-first search
	 * 
	 * @param allvotes a list of all votes for a particular bid.
	 * @return a consensus-group with maximum power, so there is no other
	 *         consensus group that has more power (but there may be other
	 *         groups with same power). returns empty set if no consensus found
	 */
	protected Set<PartyId> getMaxPowerGroupBreathFirst(Set<Vote> allvotes) {
		/**
		 * queue that keeps the options for the breath-first search sorted from
		 * highest to lowest power of the voters.
		 */
		PriorityQueue<Set<Vote>> queue = new PriorityQueue<>(
				new Comparator<Set<Vote>>() {
					@Override
					public int compare(Set<Vote> vs1, Set<Vote> vs2) {
						Set<PartyId> parties1 = vs1.stream()
								.map(vote -> vote.getActor())
								.collect(Collectors.toSet());
						Set<PartyId> parties2 = vs2.stream()
								.map(vote -> vote.getActor())
								.collect(Collectors.toSet());
						return getTotalPower(parties1)
								.compareTo(getTotalPower(parties2));
					}
				});
		queue.add(allvotes);

		while (!queue.isEmpty()) {
			// try one with the highest power
			Set<Vote> vs = queue.poll();
			if (isViable(vs))
				return getParties(vs);
			if (vs.size() > 1) {
				// vote can be splitted. push all new subsets
				for (Vote vote : vs) {
					Set<Vote> subvotes = new HashSet<>(vs);
					subvotes.remove(vote);
					// avoid duplicate work. Since this has a removed party,
					// and assuming power of each party >=1,
					// we can not already have checked subvotes
					if (!queue.contains(subvotes))
						queue.add(subvotes);
				}
			}

		}
		return Collections.emptySet();
	}

	/**
	 * 
	 * @param votes a list of all votes for a particular bid. The parties that
	 *              placed the votes are called 'voters'
	 * @return a consensus-group with maximum power, so there is no other
	 *         consensus group that has more power (but there may be other
	 *         groups with same power). returns empty set if no consensus found
	 */
	protected Set<PartyId> getMaxPowerGroup(Set<Vote> votes) {
		Set<PartyId> maxgroup = Collections.emptySet();
		Integer maxpower = 0;

		for (Set<PartyId> parties : getConsensusGroups(votes)) {
			Integer power = getTotalPower(parties);
			if (power > maxpower) {
				maxgroup = parties;
				maxpower = power;
			}
		}
		return maxgroup;
	}

	/**
	 * 
	 * @param votes a set of {@link Votes} for one single {@link Bid}. It is
	 *              assumed all votes are by different parties.
	 * @return all subsets sets of parties that create a viable consensus vote.
	 *         A viable consensus voteset VS is a subset of the votes, such that
	 *         <ol>
	 *         <li>|VS| &ge; 2
	 *         <li>TP = The sum of the powers of the parties in VS
	 *         <li>forall v in VS: v_minpower &le; TP &le; v_maxpower
	 *         </ol>
	 *         Notice that this algorithm works in exponential space
	 *         (=expensive) and not suited for large numbers of parties. Should
	 *         work fine for up to 10 parties.
	 */
	protected Set<Set<PartyId>> getConsensusGroups(Set<Vote> votes) {
		Set<Set<PartyId>> groups = new HashSet<>();

		PowerSet<Vote> allVotePermutations = new PowerSet<Vote>(
				new FixedList<Vote>(votes));
		// check each possible subset.
		// TODO stream this
		for (ImmutableList<Vote> voteList : allVotePermutations) {
			if (isViable(voteList)) {
				groups.add(getParties(voteList));
			}
		}
		return groups;
	}

	/**
	 * 
	 * @param voteList a list of votes for a single bid.
	 * @return true if this votelist satisfies the requirements of all
	 *         participants and there are &ge; 2 parties
	 */
	private boolean isViable(Iterable<Vote> voteList) {
		Set<PartyId> parties = getParties(voteList);
		Integer totalpower = getTotalPower(parties);
		return parties.size() >= 2 && totalpower >= getMinPower(voteList)
				&& totalpower <= getMaxPower(voteList);
	}

	/**
	 * @param voteList list of all {@link Vote}s
	 * @return all parties that did a vote
	 */
	public static Set<PartyId> getParties(Iterable<Vote> voteList) {
		Set<PartyId> parties = new HashSet<>();
		for (Vote vote : voteList) {
			parties.add(vote.getActor());
		}
		return parties;
	}

	/**
	 * 
	 * @param votes a list of {@link Vote}s on one Bid. All parties must be
	 *              known
	 * @return the minimum voting power needed for this group, so that all can
	 *         accept. This is equal to the max of the vote.getMinPower
	 */
	private Integer getMinPower(Iterable<Vote> votes) {
		int max = 0;
		for (Vote vote : votes) {
			max = Math.max(max, vote.getMinPower());
		}
		return max;
	}

	/**
	 * 
	 * @param votes a list of {@link Vote}s on one Bid. All parties must be
	 *              known
	 * @return the maximum voting power needed for this group, so that all can
	 *         accept. This is equal to the min of the vote.getMaxPower
	 */
	private Integer getMaxPower(Iterable<Vote> votes) {
		int min = Integer.MAX_VALUE;
		for (Vote vote : votes) {
			min = Math.min(min, vote.getMaxPower());
		}
		return min;
	}

	private Map<Bid, Set<Vote>> getAllBids1() {
		Map<Bid, Set<Vote>> bids = new HashMap<>();
		for (Votes votes : allvotes.values()) {
			for (Vote vote : votes.getVotes()) {
				Bid bid = vote.getBid();
				Set<Vote> list = bids.get(bid);
				if (list == null)
					list = new HashSet<Vote>();
				list.add(vote);
				bids.put(bid, list);
			}
		}
		return bids;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((allvotes == null) ? 0 : allvotes.hashCode());
		result = prime * result + ((powers == null) ? 0 : powers.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CollectedVotes other = (CollectedVotes) obj;
		if (allvotes == null) {
			if (other.allvotes != null)
				return false;
		} else if (!allvotes.equals(other.allvotes))
			return false;
		if (powers == null) {
			if (other.powers != null)
				return false;
		} else if (!powers.equals(other.powers))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "CollectedVotes[" + allvotes + "," + powers + "]";
	}

}
