package geniusweb.voting.votingevaluators;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;

import geniusweb.actions.PartyId;
import geniusweb.inform.Agreements;
import geniusweb.issuevalue.Bid;
import geniusweb.voting.CollectedVotes;
import geniusweb.voting.VotingEvaluator;

/**
 * {@link #getAgreements()} collects all possible {@link Agreements}. We're
 * finished only when {@link #getAgreements()} covers all except possibly 1
 * party.
 * 
 * The agreements is a list of {@link Bid}s and a maximum-sized list of
 * {@link PartyId}s that placed a satisfied vote for that bid. maximum-sized
 * means that there is no larger subset of votes for that bid. There may be
 * other subsets of equal size for the bid. If there are no agreements at all
 * for a bid, the bid is not placed in the map.
 * 
 */
public class LargestAgreementsLoop implements VotingEvaluator {
	@JsonIgnore
	private final CollectedVotes allVotes;

	// caching
	@JsonIgnore
	private Agreements maxAgreements = null;

	@JsonCreator
	public LargestAgreementsLoop() {
		this(new CollectedVotes(Collections.emptyMap(),
				Collections.emptyMap()));
	}

	protected LargestAgreementsLoop(CollectedVotes votes) {
		this.allVotes = votes;
	}

	@Override
	public Agreements getAgreements() {
		if (maxAgreements == null)
			maxAgreements = collectAgreements();
		return maxAgreements;
	}

	@Override
	public boolean isFinished() {
		HashSet<PartyId> parties = new HashSet<PartyId>(
				allVotes.getVotes().keySet());
		parties.removeAll(getAgreements().getMap().keySet());
		return parties.size() < 2;
	}

	@Override
	public VotingEvaluator create(CollectedVotes votes) {
		return new LargestAgreementsLoop(votes);
	}

	/**
	 * @return the agreements that were reached from the given votes, using the
	 *         greedy algorithm picking the largets votesets.
	 */
	protected Agreements collectAgreements() {
		CollectedVotes remainingvotes = allVotes;
		Agreements newagreements = new Agreements();
		while (true) {
			Map<Bid, Set<PartyId>> agrees = remainingvotes.getMaxAgreements();
			if (agrees.isEmpty())
				break;
			// find the bid with max group power
			Bid maxbid = agrees.keySet().stream()
					.max(Comparator.comparingInt(
							bid -> allVotes.getTotalPower(agrees.get(bid))))
					.get();
			Set<PartyId> maxparties = agrees.get(maxbid);
			newagreements = newagreements
					.with(new Agreements(maxbid, maxparties));
			remainingvotes = remainingvotes.without(maxparties);
		}
		return newagreements;
	}

	/**
	 * WARNING you must manually set prime=37 here to fix collision with
	 * {@link LargestAgreement}.
	 */
	@Override
	public int hashCode() {
		final int prime = 37;
		int result = 1;
		result = prime * result
				+ ((allVotes == null) ? 0 : allVotes.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		LargestAgreementsLoop other = (LargestAgreementsLoop) obj;
		if (allVotes == null) {
			if (other.allVotes != null)
				return false;
		} else if (!allVotes.equals(other.allVotes))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "LargestAgreementsLoop";
	}
}
